import {
  ClassType,
  Metadata,
  PropertyKeyType,
  TypeUtil,
} from '@overflow/commons/core/type';

import {AccessibleObject} from './accessible_object';
import {Annotation} from './annotation';
import {systemClassRegistry} from './class_registry';
import {Constructor} from './constructor';
import {Field} from './field';
import {Method} from './method';



export class Class extends AccessibleObject {
  private _clazzType: ClassType;
  private _constructor: Constructor;
  private _fields: Map<PropertyKeyType, Field>;
  private _methods: Map<PropertyKeyType, Method>;

  /**
   * forClass
   */
  public static forClassType(clazzType: ClassType): Class | undefined {
    return systemClassRegistry.get(clazzType);
  }

  /**
   * _defineClass
   */
  public static _defineClass(clazzType: ClassType): Class {
    let clazz: Class = Class.forClassType(clazzType);
    if (undefined === clazz) {
      clazz = new Class(clazzType);
      systemClassRegistry.set(clazzType, clazz);
    }

    if (null === clazz._constructor) {
      let parameterTypes = Metadata.getOwnParamTypes(clazzType);
      if (undefined !== parameterTypes) {
        clazz._constructor = new Constructor(clazz, parameterTypes);
      }
    }

    return clazz;
  }

  private constructor(clazzType: ClassType) {
    super();
    this._clazzType = clazzType;
    this._fields = new Map();
    this._methods = new Map();
  }

  /**
   * _defineField
   */
  public _defineConstructor(parameterTypes: any[]): Constructor {
    let cons: Constructor = this._constructor;
    if (undefined === cons) {
      cons = new Constructor(this, parameterTypes);
      this._constructor = cons;
    }
    return cons;
  }

  /**
   * _defineField
   */
  public _defineField(propertyKey: PropertyKeyType, propertyType: any): Field {
    let field: Field = this._fields.get(propertyKey);
    if (undefined === field) {
      field = new Field(this, propertyKey, propertyType);
      this._fields.set(propertyKey, field);
    }
    return field;
  }

  /**
   * _defineMethod
   */
  public _defineMethod(propertyKey: PropertyKeyType, parameterTypes: any[], returnType: any): Method {
    let method: Method = this._methods.get(propertyKey);
    if (undefined === method) {
      method = new Method(this, propertyKey, parameterTypes, returnType);
      this._methods.set(propertyKey, method);
    }
    return method;
  }

  public getType(): ClassType {
    return this._clazzType;
  }

  public getConstructor(): Constructor {
    if (undefined === this._constructor) {
      this._constructor = new Constructor(this, undefined);
    }
    return this._constructor;
  }

  public getOwnField(propertyKey: PropertyKeyType): Field | undefined {
    return this._fields.get(propertyKey);
  }

  public getOwnFields(): Map<PropertyKeyType, Field> {
    return this._fields;
  }

  public getField(propertyKey: PropertyKeyType): Field | undefined {
    let fields = this.getFields();

    return fields.get(propertyKey);
  }

  public getFields(): Map<PropertyKeyType, Field> {
    let fields: Map<PropertyKeyType, Field> = new Map();

    let clazzTypes = TypeUtil.ancestorsOf(this._clazzType);
    if (null === clazzTypes || 0 === clazzTypes.length) {
      return fields;
    }
    for (let i = 0; i < clazzTypes.length; i++) {
      const tClazzType = clazzTypes[i];
      let tClazz = Class.forClassType(tClazzType);
      if (undefined === tClazz) {
        continue;
      }

      tClazz.getOwnFields().forEach((value: Field, key: PropertyKeyType, map: Map<PropertyKeyType, Field>): void => {
        fields.set(key, value);
      });
    }

    return fields;
  }

  public getOwnMethod(propertyKey: PropertyKeyType): Method | undefined {
    return this._methods.get(propertyKey);
  }

  public getOwnMethods(): Map<PropertyKeyType, Method> {
    return this._methods;
  }

  public getMethod(propertyKey: PropertyKeyType): Method | undefined {
    let methods = this.getMethods();

    return methods.get(propertyKey);
  }

  public getMethods(): Map<PropertyKeyType, Method> {
    let methods: Map<PropertyKeyType, Method> = new Map();

    let clazzTypes = TypeUtil.ancestorsOf(this._clazzType);
    if (null === clazzTypes || 0 === clazzTypes.length) {
      return methods;
    }
    for (let i = 0; i < clazzTypes.length; i++) {
      const tClazzType = clazzTypes[i];
      let tClazz = Class.forClassType(tClazzType);
      if (undefined === tClazz) {
        continue;
      }

      tClazz.getOwnMethods().forEach((value: Method, key: PropertyKeyType, map: Map<PropertyKeyType, Method>): void => {
        methods.set(key, value);
      });
    }

    return methods;
  }

  public getAnnotation<AnnotationType extends Annotation>(annotationClass: ClassType<AnnotationType>): AnnotationType | undefined {
    let annotations = this.getAnnotations();

    return <AnnotationType>annotations.get(annotationClass);
  }

  public getAnnotations(): Map<ClassType, Annotation> {
    let annotations: Map<ClassType, Annotation> = new Map();

    let clazzTypes = TypeUtil.ancestorsOf(this._clazzType);
    if (null === clazzTypes || 0 === clazzTypes.length) {
      return annotations;
    }
    for (let i = 0; i < clazzTypes.length; i++) {
      const tClazzType = clazzTypes[i];
      let tClazz = Class.forClassType(tClazzType);
      if (undefined === tClazz) {
        continue;
      }

      tClazz.getOwnAnnotations().forEach((value: Annotation, key: ClassType, map: Map<ClassType, Annotation>): void => {
        annotations.set(key, value);
      });
    }

    return annotations;
  }


  public getName(): string {
    return this._clazzType.name;
  }
}
